/*
	MIT License

	Copyright (c) 2023 RealTimeChris

	Permission is hereby granted, free of charge, to any person obtaining a copy of this
	software and associated documentation files (the "Software"), to deal in the Software
	without restriction, including without limitation the rights to use, copy, modify, merge,
	publish, distribute, sublicense, and/or sell copies of the Software, and to permit
	persons to whom the Software is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or
	substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
	INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
	PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
	FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
	OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	DEALINGS IN THE SOFTWARE.
*/
/// https://github.com/RealTimeChris/jsonifier
/// Feb 20, 2023
#pragma once

#include <jsonifier/TypeEntities.hpp>
#include <jsonifier/StringView.hpp>
#include <jsonifier/Reflection.hpp>
#include <jsonifier/Derailleur.hpp>
#include <jsonifier/Tuple.hpp>
#include <algorithm>
#include <numeric>

namespace jsonifier_internal {

	struct xoshiro256 {
		size_t state[4]{};

		constexpr xoshiro256() noexcept {
			constexpr auto x   = 7185499250578500046ull >> 12ull;
			constexpr auto x01 = x ^ x << 25ull;
			constexpr auto x02 = x01 ^ x01 >> 27ull;
			size_t s		   = x02 * 0x2545F4914F6CDD1Dull;
			for (uint64_t y = 0; y < 4; ++y) {
				state[y] = splitmix64(s);
			}
		}

		constexpr size_t operator()() noexcept {
			const size_t result = rotl(state[1ull] * 5ull, 7ull) * 9ull;

			const size_t t = state[1ull] << 17ull;

			state[2ull] ^= state[0ull];
			state[3ull] ^= state[1ull];
			state[1ull] ^= state[2ull];
			state[0ull] ^= state[3ull];

			state[2ull] ^= t;

			state[3ull] = rotl(state[3ull], 45ull);

			return result;
		}

	  protected:
		constexpr size_t rotl(const size_t x, size_t k) const noexcept {
			return (x << k) | (x >> (64ull - k));
		}

		constexpr size_t splitmix64(size_t& seed64) const noexcept {
			size_t result = seed64 += 0x9E3779B97F4A7C15ull;
			result		  = (result ^ (result >> 30ull)) * 0xBF58476D1CE4E5B9ull;
			result		  = (result ^ (result >> 27ull)) * 0x94D049BB133111EBull;
			return result ^ (result >> 31ull);
		}
	};

	struct tuple_reference {
		size_t stringLength{};
		size_t oldIndex{};
		size_t newIndex{};
	};

	struct string_length_counts {
		size_t length{};
		size_t count{};
	};

	template<size_t currentIndex = 0, size_t maxIndex, typename tuple_type>
	JSONIFIER_ALWAYS_INLINE constexpr auto collectTupleReferencesInternal(const tuple_type& tuple, std::array<tuple_reference, maxIndex>& arrayOfRefs) noexcept {
		if constexpr (currentIndex < maxIndex) {
			tuple_reference values{};
			values.oldIndex			  = currentIndex;
			values.stringLength		  = std::get<currentIndex>(tuple).view().size();
			arrayOfRefs[currentIndex] = values;
			return collectTupleReferencesInternal<currentIndex + 1>(tuple, arrayOfRefs);
		} else {
			return arrayOfRefs;
		}
	}

	template<typename tuple_type> JSONIFIER_ALWAYS_INLINE constexpr auto collectTupleReferences(const tuple_type& tuple) noexcept {
		constexpr auto maxIndex = std::tuple_size_v<unwrap_t<tuple_type>>;
		std::array<tuple_reference, maxIndex> arrayOfRefs{};
		return collectTupleReferencesInternal<0, maxIndex>(tuple, arrayOfRefs);
	}

	template<size_t size> JSONIFIER_ALWAYS_INLINE constexpr auto bubbleSort(std::array<tuple_reference, size>& arrayOfRefs) noexcept {
		for (size_t i = 0; i < size - 1; ++i) {
			for (size_t j = 0; j < size - i - 1; ++j) {
				if (arrayOfRefs[j].stringLength > arrayOfRefs[j + 1].stringLength) {
					std::swap(arrayOfRefs[j], arrayOfRefs[j + 1]);
				}
			}
		}
		for (size_t i = 0; i < size; ++i) {
			arrayOfRefs[i].newIndex = i;
		}
	}

	template<typename value_type> JSONIFIER_ALWAYS_INLINE constexpr auto sortTupleReferences() noexcept {
		constexpr auto& tuple = jsonifier::core<unwrap_t<value_type>>::parseValue.parseValue;
		auto collectedRefs	  = collectTupleReferences(tuple);
		bubbleSort(collectedRefs);
		return collectedRefs;
	}

	template<size_t start, size_t end, size_t... indices> JSONIFIER_ALWAYS_INLINE constexpr auto generateCustomIndexSequence(std::index_sequence<indices...>) noexcept {
		return std::index_sequence<(start + indices)...>{};
	}

	template<size_t start, size_t end> struct custom_index_sequence_generator;

	template<size_t start, size_t end> struct custom_index_sequence_generator {
		using type = decltype(generateCustomIndexSequence<start, end>(std::make_index_sequence<end - start + 1>{}));
	};

	template<size_t start, size_t end> using custom_index_sequence = typename custom_index_sequence_generator<start, end>::type;

	template<size_t size> JSONIFIER_ALWAYS_INLINE constexpr auto countUniqueStringLengths(const std::array<tuple_reference, size>& arrayOfRefs) noexcept {
		size_t currentLength{};
		size_t currentCount{};
		for (size_t x = 0; x < size; ++x) {
			if (arrayOfRefs[x].stringLength != currentLength) {
				currentLength = arrayOfRefs[x].stringLength;
				++currentCount;
			}
		}
		return currentCount;
	}

	template<size_t counts, size_t size> JSONIFIER_ALWAYS_INLINE constexpr auto collectUniqueStringLengths(const std::array<tuple_reference, size>& arrayOfRefs) noexcept {
		std::array<string_length_counts, counts> returnValues{};
		size_t currentIndex	 = 0;
		size_t currentLength = 0;

		for (size_t x = 0; x < arrayOfRefs.size(); ++x) {
			auto keyLength = arrayOfRefs[x].stringLength;
			if (currentLength != keyLength && currentIndex < counts) {
				currentLength					  = keyLength;
				returnValues[currentIndex].length = keyLength;
				returnValues[currentIndex].count  = 1;
				++currentIndex;
			} else {
				if (currentIndex - 1 < returnValues.size()) {
					++returnValues[currentIndex - 1].count;
				}
			}
		}

		return returnValues;
	}

	template<uint64_t inputIndex, const auto& stringLengths> JSONIFIER_ALWAYS_INLINE constexpr auto getCurrentStringLength() noexcept {
		for (size_t x = 0; x < stringLengths.size(); ++x) {
			if (inputIndex == x) {
				return stringLengths[x].length;
			}
		}
		return std::numeric_limits<size_t>::max();
	}

	template<size_t subTupleIndex, const auto& stringLengths> JSONIFIER_ALWAYS_INLINE constexpr auto getCurrentStartingIndex() noexcept {
		size_t currentStartingIndex{};
		for (size_t x = 0; x < subTupleIndex; ++x) {
			currentStartingIndex += stringLengths[x].count;
		}
		return currentStartingIndex;
	}

	template<typename value_type> inline static constexpr auto sortedTupleReferences{ sortTupleReferences<unwrap_t<value_type>>() };
	template<typename value_type> inline static constexpr auto uniqueStringLengthCount{ countUniqueStringLengths(sortedTupleReferences<value_type>) };
	template<typename value_type>
	inline static constexpr auto uniqueStringLengths{ collectUniqueStringLengths<uniqueStringLengthCount<value_type>>(sortedTupleReferences<value_type>) };

	template<typename value_type, size_t... indices> JSONIFIER_ALWAYS_INLINE constexpr auto createGroupedSubTuple(std::index_sequence<indices...>) noexcept {
		constexpr auto& oldTuple  = jsonifier::concepts::coreV<value_type>;
		constexpr auto& tupleRefs = sortedTupleReferences<value_type>;
		return std::make_tuple(std::get<tupleRefs[indices].oldIndex>(oldTuple)...);
	}

	template<typename value_type, size_t startIndex, size_t count> JSONIFIER_ALWAYS_INLINE constexpr auto createGroupedTuple() noexcept {
		return createGroupedSubTuple<value_type>(custom_index_sequence<startIndex, (startIndex + count) - 1>{});
	}

	template<typename value_type, size_t... indices> JSONIFIER_ALWAYS_INLINE constexpr auto createNewTupleInternal(std::index_sequence<indices...>) noexcept {
		constexpr auto& stringLengths = uniqueStringLengths<value_type>;
		return std::make_tuple(createGroupedTuple<value_type, getCurrentStartingIndex<indices, stringLengths>(), stringLengths[indices].count>()...);
	}

	template<typename value_type> JSONIFIER_ALWAYS_INLINE constexpr auto createNewTuple() noexcept {
		constexpr auto& tupleRefs = sortedTupleReferences<value_type>;
		return createNewTupleInternal<value_type>(std::make_index_sequence<countUniqueStringLengths(tupleRefs)>{});
	}

	template<typename value_type> inline static constexpr auto final_tuple_static_data = createNewTuple<unwrap_t<value_type>>();

	template<typename value_type, size_t subTupleIndex, size_t I> JSONIFIER_ALWAYS_INLINE constexpr const jsonifier::string_view& getKey() noexcept {
		constexpr auto& subTuple = std::get<subTupleIndex>(final_tuple_static_data<value_type>);
		return std::get<I>(subTuple).view();
	}

	template<typename value_type, size_t subTupleIndex, uint64_t maxIndex, uint64_t index> JSONIFIER_ALWAYS_INLINE constexpr auto keyStatsInternal(key_stats_t stats) noexcept {
		if constexpr (index < maxIndex) {
			constexpr const jsonifier::string_view& key{ getKey<value_type, subTupleIndex, index>() };
			constexpr auto n{ key.size() };
			if (n < stats.minLength) {
				stats.minLength = n;
			}
			if (n > stats.maxLength) {
				stats.maxLength = n;
			}
			return keyStatsInternal<value_type, subTupleIndex, maxIndex, index + 1>(stats);
		} else {
			if constexpr (maxIndex > 0) {
				stats.lengthRange = stats.maxLength - stats.minLength;
			}
			return stats;
		}
	}

	template<typename value_type, size_t subTupleIndex> JSONIFIER_ALWAYS_INLINE constexpr auto keyStats() noexcept {
		constexpr auto& subTuple = std::get<subTupleIndex>(final_tuple_static_data<value_type>);
		constexpr auto N{ std::tuple_size_v<unwrap_t<decltype(subTuple)>> };

		return keyStatsInternal<value_type, subTupleIndex, N, 0>(key_stats_t{});
	}

	template<typename value_type01, typename value_type02>
	JSONIFIER_ALWAYS_INLINE constexpr bool contains(const value_type01* hashData, value_type02 byteToCheckFor, size_t size) noexcept {
		for (size_t x = 0; x < size; ++x) {
			if (hashData[x] == byteToCheckFor) {
				return true;
			}
		}
		return false;
	}

	constexpr std::array<size_t, 5> hashTupleMaxSizes{ 16, 32, 64, 128, 256 };

	template<typename value_type, size_t subTupleIndex, uint64_t index> struct string_compare_wrapper_non_const {
		JSONIFIER_ALWAYS_INLINE static bool op(const char* string01) noexcept {
			static constexpr auto& subTuple = std::get<subTupleIndex>(final_tuple_static_data<value_type>);
			if constexpr (index < std::tuple_size_v<unwrap_t<decltype(subTuple)>>) {
				static constexpr auto currentKey = getKey<value_type, subTupleIndex, index>();
				return compare<currentKey.size()>(currentKey.data(), string01);
			} else {
				return false;
			}
		}
	};

	template<typename value_type, size_t subTupleIndex> JSONIFIER_ALWAYS_INLINE constexpr size_t getActualSize() noexcept {
		constexpr auto& subTuple = std::get<subTupleIndex>(final_tuple_static_data<value_type>);
		using sub_tuple_t		 = unwrap_t<decltype(subTuple)>;
		return std::tuple_size_v<sub_tuple_t>;
	}

	template<template<typename, size_t, uint64_t> typename function_wrapper, typename value_type, size_t subTupleIndex, size_t... indices>
	JSONIFIER_ALWAYS_INLINE constexpr auto generateArrayOfFunctionPtrs(std::index_sequence<indices...>) noexcept {
		return std::array<decltype(&function_wrapper<value_type, subTupleIndex, 0>::op), sizeof...(indices)>{ &function_wrapper<value_type, subTupleIndex, indices>::op... };
	}

	template<typename key_type_new, typename value_type, size_t subTupleIndexNew, size_t maxSizeIndexNew> struct sub_tuple_construction_data {
		JSONIFIER_ALIGN std::array<size_t, hashTupleMaxSizes[maxSizeIndexNew] + 1> indices{};
		size_t actualCount{ getActualSize<value_type, subTupleIndexNew>() };
		size_t storageSize{ hashTupleMaxSizes[maxSizeIndexNew] };
		size_t subTupleIndex{ subTupleIndexNew };
		size_t uniqueIndex{};
		size_t seed{};

		JSONIFIER_ALWAYS_INLINE constexpr sub_tuple_construction_data() noexcept = default;
	};

	template<typename key_type, typename value_type, size_t subTupleIndex, template<typename, typename, size_t, size_t> typename sub_tuple_construction_data_type>
	using sub_tuple_construction_data_variant = std::variant<sub_tuple_construction_data_type<key_type, value_type, subTupleIndex, 0>,
		sub_tuple_construction_data_type<key_type, value_type, subTupleIndex, 1>, sub_tuple_construction_data_type<key_type, value_type, subTupleIndex, 2>,
		sub_tuple_construction_data_type<key_type, value_type, subTupleIndex, 3>, sub_tuple_construction_data_type<key_type, value_type, subTupleIndex, 4>>;

	template<size_t size>
	JSONIFIER_ALWAYS_INLINE constexpr size_t findUniqueColumnIndex(const std::array<jsonifier::string_view, size>& strings, const key_stats_t& keyStats) noexcept {
		constexpr size_t alphabetSize = 256;

		for (size_t index = 0; index < keyStats.maxLength; ++index) {
			bool allDifferent = true;
			std::array<bool, alphabetSize> seen{};

			for (const auto& str: strings) {
				char c = str[index % str.size()];
				if (seen[static_cast<unsigned char>(c)]) {
					allDifferent = false;
					break;
				}
				seen[static_cast<unsigned char>(c)] = true;
			}

			if (allDifferent) {
				return index;
			}
		}

		return keyStats.minLength;
	}

	template<typename key_type, typename value_type, size_t subTupleIndexNew>
	JSONIFIER_ALWAYS_INLINE constexpr auto collectMinimalCharSubTupleData(const std::array<key_type, getActualSize<value_type, subTupleIndexNew>()>& pairsNew) noexcept {
		auto keyStatsVal = keyStats<value_type, subTupleIndexNew>();
		xoshiro256 prng{};
		const auto uniqueIndex = findUniqueColumnIndex(pairsNew, keyStatsVal);
		auto constructForGivenStringLength =
			[&](const auto maxSizeIndex, auto&& constructForGivenStringLength,
				size_t uniqueIndex) mutable -> sub_tuple_construction_data_variant<key_type, value_type, subTupleIndexNew, sub_tuple_construction_data> {
			constexpr size_t storageSize = hashTupleMaxSizes[maxSizeIndex];
			sub_tuple_construction_data<key_type, value_type, subTupleIndexNew, maxSizeIndex> returnValues{};
			returnValues.uniqueIndex = uniqueIndex;
			bool collided{};
			for (size_t x = 0; x < 2; ++x) {
				std::fill(returnValues.indices.begin(), returnValues.indices.end(), returnValues.indices.size() - 1);
				returnValues.seed = prng();
				collided		  = false;
				for (size_t y = 0; y < getActualSize<value_type, subTupleIndexNew>(); ++y) {
					const auto hash = (returnValues.seed ^ pairsNew[y].data()[uniqueIndex]);
					const auto slot = hash % storageSize;
					if (returnValues.indices[slot] != returnValues.indices.size() - 1) {
						collided = true;
						break;
					}
					returnValues.indices[slot] = y;
				}
				if (!collided) {
					break;
				}
			}
			if (collided) {
				if constexpr (maxSizeIndex < std::size(hashTupleMaxSizes) - 1) {
					return sub_tuple_construction_data_variant<key_type, value_type, subTupleIndexNew, sub_tuple_construction_data>{ constructForGivenStringLength(
						std::integral_constant<size_t, maxSizeIndex + 1>{}, constructForGivenStringLength, uniqueIndex) };
				}
				returnValues.uniqueIndex = std::numeric_limits<size_t>::max();
				return returnValues;
			}
			return returnValues;
		};

		return constructForGivenStringLength(std::integral_constant<size_t, 0>{}, constructForGivenStringLength, uniqueIndex);
	}

	template<typename value_type, size_t subTupleIndex, size_t... I>
	JSONIFIER_ALWAYS_INLINE constexpr auto generateSubTupleConstructionDataImpl(std::index_sequence<I...>) noexcept {
		constexpr auto& tuple	 = std::get<subTupleIndex>(final_tuple_static_data<value_type>);
		using tuple_type		 = unwrap_t<decltype(tuple)>;
		constexpr auto tupleSize = std::tuple_size_v<tuple_type>;

		if constexpr (tupleSize == 0) {
			return nullptr;
		} else {
			return collectMinimalCharSubTupleData<jsonifier::string_view, value_type, subTupleIndex>({ getKey<value_type, subTupleIndex, I>()... });
		}
	}

	template<typename value_type, size_t subTupleIndex> JSONIFIER_ALWAYS_INLINE constexpr auto generateSubTupleConstructionData() noexcept {
		constexpr auto tupleSize = std::tuple_size_v<unwrap_t<decltype(std::get<subTupleIndex>(final_tuple_static_data<value_type>))>>;
		constexpr auto mapNew	 = generateSubTupleConstructionDataImpl<value_type, subTupleIndex>(std::make_index_sequence<tupleSize>{});
		constexpr auto newIndex	 = mapNew.index();
		return std::get<newIndex>(mapNew);
	}

	template<typename value_type, size_t subTupleIndex>
	inline static constexpr auto sub_tuple_construction_static_data{ generateSubTupleConstructionData<unwrap_t<value_type>, subTupleIndex>() };

	template<typename value_type> struct hash_tuple;

	template<typename value_type, size_t subTupleIndexNew> struct minimal_char_sub_tuple {
		inline static constexpr auto& constructionData{ sub_tuple_construction_static_data<value_type, subTupleIndexNew> };
		using size_type	   = size_t;

		std::array<size_type, constructionData.storageSize + 1> indices{ constructionData.indices };
		size_t uniqueIndex{ constructionData.uniqueIndex };
		size_t storageSize{ constructionData.storageSize };
		size_t seed{ constructionData.seed };

		JSONIFIER_ALWAYS_INLINE constexpr minimal_char_sub_tuple() noexcept = default;

		template<const auto& functionPtrs, typename... arg_types> JSONIFIER_ALWAYS_INLINE auto find(const char* iter, arg_types&&... args) const noexcept {
			static constexpr auto stringCompareFunctions{ generateArrayOfFunctionPtrs<string_compare_wrapper_non_const, value_type, subTupleIndexNew>(
				std::make_index_sequence<constructionData.storageSize + 1>{}) };
			const auto hash		  = (seed ^ iter[uniqueIndex]);
			const auto finalIndex = hash % storageSize;
			if (stringCompareFunctions[indices[finalIndex]](iter)) {
				functionPtrs[indices[finalIndex]](std::forward<arg_types>(args)...);
				return true;
			} else {
				return false;
			}
		}
	};

	template<size_t inputIndex, const auto& stringLengths> JSONIFIER_ALWAYS_INLINE constexpr auto getCurrentSubTupleIndex() noexcept {
		for (size_t x = 0; x < stringLengths.size(); ++x) {
			if (inputIndex == stringLengths[x].length) {
				return x;
			}
		}
		return std::numeric_limits<size_t>::max();
	}

	template<const auto& function, typename value_type, size_t uniqueIndex, typename... arg_types>
	JSONIFIER_ALWAYS_INLINE constexpr auto collectSubTuple(arg_types&&... args) noexcept {
		using tuple_type	   = unwrap_t<decltype(final_tuple_static_data<value_type>)>;
		constexpr size_t index = getCurrentSubTupleIndex<uniqueIndex, uniqueStringLengths<value_type>>();
		if constexpr (index < std::tuple_size_v<tuple_type>) {
			return function(std::integral_constant<size_t, index>{}, std::forward<arg_types>(args)...);
		} else {
			return false;
		}
	};

	template<const auto& function, typename value_type, size_t index, typename... arg_types> using collect_sub_tuple_ptr =
		decltype(&collectSubTuple<function, value_type, index, arg_types...>);

	template<const auto& function, typename value_type, typename... arg_types, size_t... indices>
	JSONIFIER_ALWAYS_INLINE constexpr auto generateCollectSubTuplePtrArrayInternal(std::index_sequence<indices...>) noexcept {
		return std::array<collect_sub_tuple_ptr<function, value_type, 0, arg_types...>, sizeof...(indices)>{ { &collectSubTuple<function, value_type, indices, arg_types...>... } };
	}

	template<const auto& function, typename value_type, typename... arg_types> JSONIFIER_ALWAYS_INLINE constexpr auto generateCollectSubTuplePtrArray() noexcept {
		return generateCollectSubTuplePtrArrayInternal<function, value_type, arg_types...>(std::make_index_sequence<512>{});
	}

	template<typename value_type> struct hash_tuple {
		using size_type = size_t;

		JSONIFIER_ALWAYS_INLINE constexpr hash_tuple() noexcept = default;

		template<size_t subTupleIndex, const auto& subTupleFunctionPtrArray, typename... arg_types>
		JSONIFIER_ALWAYS_INLINE static auto find(const char* ptr01, arg_types&&... args) noexcept {
			static constexpr auto subTuple{ minimal_char_sub_tuple<value_type, subTupleIndex>{} };
			return subTuple.template find<subTupleFunctionPtrArray>(ptr01, std::forward<arg_types>(args)...);
		}

		template<const auto& function, typename... arg_types> JSONIFIER_ALWAYS_INLINE static auto find(size_type uniqueIndex, arg_types&&... args) noexcept {
			static constexpr auto getSubTuplePtrArray = generateCollectSubTuplePtrArray<function, unwrap_t<value_type>, arg_types...>();
			return getSubTuplePtrArray[uniqueIndex](std::forward<arg_types>(args)...);
		}
	};

}